import sys
import time
import numpy as np
import math

from mpl_toolkits.mplot3d import axes3d
import mpl_toolkits.mplot3d.art3d as art3d
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.patches import Circle, PathPatch, Ellipse

import roboticstoolbox as rtb
from spatialmath import *
from spatialmath.base import *

def plotAxes(m, ax, lw=1, ln=1):
    ax.plot3D([m[0,3],m[0,0]*ln+m[0,3]], [m[1,3],m[1,0]*ln+m[1,3]], [m[2,3],m[2,0]*ln+m[2,3]], 'red', linewidth=lw)
    ax.plot3D([m[0,3],m[0,1]*ln+m[0,3]], [m[1,3],m[1,1]*ln+m[1,3]], [m[2,3],m[2,1]*ln+m[2,3]], 'green', linewidth=lw)
    ax.plot3D([m[0,3],m[0,2]*ln+m[0,3]], [m[1,3],m[1,2]*ln+m[1,3]], [m[2,3],m[2,2]*ln+m[2,3]], 'blue', linewidth=lw)

def setPlot():
    fig = plt.figure()
    ax  = fig.add_subplot(111, projection='3d')
    ax.set_zlim(-1, 2)  
    ax.set_xlim(-1, 2)  
    ax.set_ylim(-1, 2)
    
    plotAxes(transl(0,0,0), ax, lw=1, ln=2)
    return ax

def velocity(q1, q2, joint, j_type):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    plt.axis('equal')

    ax.set_xlim([-3.5, 3.5])
    ax.set_ylim([-3.5, 3.5])

    # load some test data for demonstration and plot a wireframe
    ax.plot([0, 1], [0, 0], [0, 0], c='red')
    ax.plot([0, 0], [0, 1], [0, 0], c='green')

    link_1, = ax.plot([0, 1], [0, 0], lw=6, c='red')
    link_2, = ax.plot([1, 2], [0, 0], lw=6, c='green')

    l1 = 1
    l2 = 1
    
    def update(i, q1, q2, val1, val2):
        val2 = np.radians(val2)
        imp = 90/val1
        it = np.radians(-(abs(i-90)-90)/imp) 
        point1 = np.array([l1*np.cos(it+q1),l1*np.sin(it+q1)])
        point2 = np.array([point1[0]+l2*np.cos(it-(abs(i-90)-90)*val2/90+q1+q2),
                           point1[1]+l2*np.sin(it-(abs(i-90)-90)*val2/90+q1+q2)])

        link_1.set_data([0, point1[0]],[0, point1[1]])
        link_2.set_data([point1[0], point2[0]],[point1[1], point2[1]])
    
    def update_p1(i, q1, q2, val1, val2):
        point1 = np.array([(l1+i/180/5)*np.cos(q1),(l1+i/180/5)*np.sin(q1)])
        point2 = np.array([point1[0]+l2*np.cos(q1+q2),point1[1]+l2*np.sin(q1+q2)])

        link_1.set_data([0, point1[0]],[0, point1[1]])
        link_2.set_data([point1[0], point2[0]],[point1[1], point2[1]])
    
    
    def update_p2(i, q1, q2, val1, val2):
        point1 = np.array([l1*np.cos(q1),l1*np.sin(q1)])
        point2 = np.array([point1[0]+(l2+i/180/5)*np.cos(q1+q2),point1[1]+(l2+i/180/5)*np.sin(q1+q2)])

        link_1.set_data([0, point1[0]],[0, point1[1]])
        link_2.set_data([point1[0], point2[0]],[point1[1], point2[1]])
    
    if (joint =='q1'):
        val1, val2 = 6, 0.01  #degrees 
    else:
        val1, val2 = 0.01, 6
    
    if (j_type == 'prismatic' and joint == 'q1' ):
        line_ani = animation.FuncAnimation(fig, update_p1, 180, interval=1, blit=False, fargs=(q1, q2, val1, val2,), repeat = True )
        return line_ani

    elif (j_type == 'prismatic' and joint == 'q2'):
        line_ani = animation.FuncAnimation(fig, update_p2, 180, interval=1, blit=False, fargs=(q1, q2, val1, val2,), repeat = True )
        return line_ani

    else:
        line_ani = animation.FuncAnimation(fig, update, 180, interval=1, blit=False, fargs=(q1, q2, val1, val2,), repeat = True )
        return line_ani
    
def velocity_ell(q1, q2):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    plt.axis('equal')

    ax.set_xlim([-3.5, 3.5])
    ax.set_ylim([-3.5, 3.5])

    # load some test data for demonstration and plot a wireframe
    ax.plot([0, 1], [0, 0], [0, 0], c='red')
    ax.plot([0, 0], [0, 1], [0, 0], c='green')

    link_1, = ax.plot([0, 0], [1, 0], lw=6, c='red')
    link_2, = ax.plot([1, 2], [0, 0], lw=6, c='green')

    el = Ellipse((2,0),1,0,angle=90,linewidth=3, fill=False)

    ax.add_patch(el)

    l1 = 1
    l2 = 1

    point1 = np.array([l1*np.cos(q1),l1*np.sin(q1)])
    point2 = np.array([point1[0]+l2*np.cos(q2+q1),point1[1]+l2*np.sin(q2+q1)])

    link_1.set_data([0, point1[0]],[0, point1[1]]);
    link_2.set_data([point1[0], point2[0]],[point1[1], point2[1]]);
    
    Jac = np.array([[-l1*np.sin(q1)-l2*np.sin(q1+q2),-l2*np.sin(q1+q2)],[l1*np.cos(q1)+l2*np.cos(q1+q2),l2*np.cos(q1+q2)]])

    A = np.matmul(Jac,np.transpose(Jac))
    w, v = np.linalg.eig(A)
    el.center = (point2[0],point2[1])
    el.height = np.sqrt(min(w))
    el.width = np.sqrt(max(w))
    el.angle = np.degrees(np.arccos(v[1][0]))
    
    plt.show()
        
        
def lim_prismatic(q, dq, val, pos):
     if (abs(q[pos]) > val):
        q[pos] = val*q[pos]/abs(q[pos])
        dq[pos] = 0

def plot_interpolation(time, s, q, dq, ddq, title, plot_der):
    if plot_der:
        fig, axs = plt.subplots(4,1)
        fig.set_figheight(10)
    else:
        fig, axs = plt.subplots(2,1)
        fig.set_figheight(5)

    axs[0].plot(time,s)
    axs[0].set_ylabel("s(t)", fontsize=20)
    axs[0].grid(True)
    axs[1].plot(time,q)
    axs[1].set_ylabel("q(t)", fontsize=20)
    axs[1].set_xlabel("time", fontsize=20)
    axs[1].grid(True)
    if plot_der:
        axs[2].plot(time,dq)
        axs[2].set_ylabel("$\dot{q}(t)$", fontsize=20)
        axs[2].set_xlabel("time", fontsize=20)
        axs[2].grid(True)
        axs[3].plot(time,ddq)
        axs[3].set_ylabel("$\ddot{q}(t)$", fontsize=20)
        axs[3].set_xlabel("time", fontsize=20)
        axs[3].grid(True)
    plt.suptitle(title, fontsize=30)
    fig.set_figwidth(15)
    plt.show()
    
def plot_car_interpolation(time, x, y, z, title):
    fig, axs = plt.subplots(3,1)
    fig.set_figheight(10)
    axs[0].plot(time,x)
    axs[0].set_ylabel("$x(t)$", fontsize=20)
    axs[0].grid(True)
    axs[1].plot(time,y)
    axs[1].set_ylabel("$y(t)$", fontsize=20)
    axs[1].grid(True)
    axs[2].plot(time,z)
    axs[2].set_ylabel("$z(t)$", fontsize=20)
    axs[2].set_xlabel("time", fontsize=20)
    axs[2].grid(True)
    plt.suptitle(title, fontsize=30)
    fig.set_figwidth(15)
    plt.show()
    
def plot_joint_interpolation(time, qs, title):
    njoints = qs.shape[1]
    fig, axs = plt.subplots(njoints,1)
    fig.set_figheight(10)
    for i in range(njoints):
        axs[i].plot(time,qs[:,i])
        axs[i].set_ylabel("$q_"+str(i)+"(t)$", fontsize=20)
        axs[i].grid(True)
    
    plt.suptitle(title, fontsize=30)
    fig.set_figwidth(15)
    plt.show()

def printProgressBar(
    fraction, prefix="", suffix="", decimals=1, length=50, fill="█", printEnd="\r"
):

    percent = ("{0:." + str(decimals) + "f}").format(fraction * 100)
    filledLength = int(length * fraction)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)
